#include <stdio.h>
#include <Tmk.h>

#include "tsp.h"

unsigned 	CurDist, PathLen;
int 		Visit[MAX_TOUR_SIZE], Path[MAX_TOUR_SIZE];
int		visitNodes;

/*
 *   recursive_solve(curr_ind)
 *
 *	We're now supposed to recursively try all possible solutions
 *	starting with the current tour.  We do this by copying the
 *	state to local variables (to avoid unneeded conflicts) and
 *	calling visit_nodes to do the actual recursive solution.
 *
 */
recursive_solve(index)
    int index;
{
    register unsigned i, j;
    register TourElement *curr = &global->Tours[index];

    CurDist = curr->prefix_weight;
    PathLen = curr->last + 1;

    for (i = 0; i < TspSize; i++) Visit[i] = 0;

    for (i = 0; i < PathLen; i++) {
	Path[i] = curr->prefix[i];
	Visit[Path[i]] = 1;
    }

    if (PathLen > TspSize) {
	fprintf(stderr, "Pathlen: %d\n", PathLen);
	fflush(stderr);
	exit(0);
    }

    if (CurDist == 0 || debugPrioQ) {
	sprintf(tour_str, "\t%d: Tour %d is ", Tmk_proc_id, index);
	j = strlen(tour_str);
	for (i = 0; i < PathLen-1; i++) {
	    sprintf((tour_str + j), "%1d - ", (int)Path[i]);
	    if ((int)Path[i] >= 10) j += 5;	/* Two digit number. */
	    else j += 4;
	}
	sprintf((tour_str + j), "%1d\n", (int)Path[i]);
	j = strlen(tour_str);
	sprintf((tour_str+j), "\t%d: Cur: %d, Min %d, Len: %d, Sz: %d.\n",
		Tmk_proc_id, CurDist, global->MinTourLen, PathLen-1, TspSize);
    }

    visit_nodes(Path[PathLen-1]);
}


/*
 *   visit_nodes()
 *
 *       Exhaustively visits each node to find Hamilton cycle.
 *       Assumes that search started at node from.
 *
 */
visit_nodes(from)
    int from;
{
    register int i;
    register int dist, last;

#ifdef	DEBUG
    visitNodes++;
#endif	/*DEBUG*/
    for (i = 1; i < TspSize; i++) {
	if (Visit[i]) continue;	/* Already visited. */
	if ((dist = global->weights[from][i]) == 0) continue; /* Not connected. */
	if (CurDist + dist > global->MinTourLen) continue; /* Path too long. */

	/* Try next node. */
	Visit[i] = 1;
	Path[PathLen++] = i;
	CurDist += dist;

	if (PathLen == TspSize) {
	    /* Visiting last node - determine if path is min length. */
	    if ((last = global->weights[i][StartNode]) != 0 &&
		(CurDist += last) < global->MinTourLen) {
		set_best(CurDist, Path);
	    }
	    CurDist -= last;
	} /* if visiting last node */
	else if (CurDist < global->MinTourLen) visit_nodes(i);	/* Visit on. */

	/* Remove current try and loop again at this level. */
	CurDist -= dist;
	PathLen--;
	Visit[i] = 0;
    }
}


/*
 *  Add up min edges connecting all unconnected vertixes (AHU p. 331-335)
 *  At some point, we may want to use MST to get a lower bound. This
 *  bound should be higher (and therefore more accurate) but will take
 *  longer to compute. 
 */
calc_bound(curr_index)
    int curr_index;
{
    register int i, j, wt, wt1, wt2;
    register TourElement *curr = &global->Tours[curr_index];

    /*
     * wt1: the value of the edge with the lowest weight from the node
     *	    we're testing to another unconnected node.
     * wt2: the value of the edge with the second lowest weight
     */

    /* if we have one unconnected node */
    if (curr->last == (TspSize - 2)) {
	for (i = 0; i < TspSize; i++) {
	    if (!(curr->conn & (1<<i))) {
		/* we have found the one unconnected node */
		curr->prefix[TspSize-1] = i;
		curr->prefix[TspSize] = StartNode;
		
		/* add edges to and from the last node */
		curr->prefix_weight+= global->weights[curr->prefix[TspSize-2]][i] +
		                      global->weights[i][curr->prefix[StartNode]];

		if (curr->prefix_weight < global->MinTourLen) {
		    /* Store our new best path and its weight. */
		    set_best(curr->prefix_weight, curr->prefix);
		}

		/* De-allocate this tour so someone else can use it */
		curr->lower_bound = BIGINT;
#ifdef	LOCK_PREFETCH
		PREFETCH2(global->TourStack, global->PrioQ);
#endif	/*LOCK_PREFETCH*/
		Tmk_lock_acquire(global->TourLock);
		global->TourStack[++global->TourStackTop] = curr_index; /* Free tour. */
		Tmk_lock_release(global->TourLock);
		return(END_TOUR);
	    }
	}
    }

    curr->mst_weight = 0;

    /*
     * Add up the lowest weights for edges connected to vertices
     * not yet used or at the ends of the current tour, and divide by two.
     * This could be tweaked quite a bit.  For example:
     *   (1) Check to make sure that using an edge would not make it
     *       impossible for a vertex to have degree two.
     *   (2) Check to make sure that the edge doesn't give some
     *       vertex degree 3.
     */

    if (curr->last != TspSize - 1) {
	for (i = 0; (i < TspSize); i++) {
	    if (curr->conn & 1<<i) continue;
	    for (j = 0, wt1 = wt2 = BIGINT; j < TspSize; j++) {
		/* Ignore j's that are not connected to i (global->weights[i][j]==0), */
		/* or that are already in the tour and aren't either the      */
		/* first or last node in the current tour.		      */
		wt = global->weights[i][j];
		if (!wt || (curr->conn&(1<<j)&&(j != curr->prefix[0]) &&
					       (j != curr->prefix[curr->last])))
		    continue;

		/* Might want to check that edges go to unused vertices */
		if (wt < wt1) {
		    wt2 = wt1;
		    wt1 = wt;
		}
		else if (wt < wt2) wt2 = wt;
	    }
  
	    /* At least three unconnected nodes? */
	    if (wt2 != BIGINT) curr->mst_weight += (wt1 + wt2) >> 1;
	    /* Exactly two unconnected nodes? */
	    else if (wt1 != BIGINT) curr->mst_weight += wt1;
	}
  	curr->mst_weight += 1;
    }

    curr->lower_bound = curr->mst_weight + curr->prefix_weight;

    return(curr_index);
}
